#!/usr/bin/python2 -u

from __future__ import absolute_import

#
# This is a trivial demo auth server, which just approves any requests
# it sees, while printing debug information to STDERR.
#
# This code is in the public domain, feel free to adapt to your needs.
#
import getopt
import json
import os
import sys


# By default we advertise only the AUTH method; ZK-AUTH is not listed.
# Roughly 50% of the time we'll offer SERVER mode as well.
CAPABILITIES = (('SERVER ' if (os.getpid() % 2 == 0) else '')
                + 'AUTH')


def Auth(domain):
    # This method simply returns a dictionary of quota values, along with
    # the shared secret in the clear. Pagekite.py takes care of challenging
    # and authenticating the user.
    #
    # Note that the quota values are mostly advisory; it is assumed that
    # accounting happens elsewhere and Pagekite.py is mostly just relaying
    # values back to the user as information.
    #
    # Exception #1: a quota_kb value of zero means the user exists but no
    # longer has access to the service (ran out of quota). Negative or
    # missing quota_kb values mean "no bandwidth quota, just grant access."
    #
    # Exception #2: The IPs-per-second rate limits are enforced by
    # pagekite.py, using the semantics of the --ratelimit_ips option.
    #
    # Valid rejection reasons that will be understood by the connector
    # include: "unauthorized", "quota", "nodays", "noquota", "noconns"
    #
    # Note that rejection reasons are ignored unless quota_kb is 0.
    #
    return {
        'secret': 'testing',  # Important! This is used for authentication.

        'quota_kb': 10240,    # If set, 0 means "out of quota"; else advisory
        'quota_days': 24,     # Advisory
        'quota_conns': 5,     # Advisory
        'reason': 'reason',   # If set, explains why user is "out of quota"

        'ips_per_sec-ips': 1,    # Inform PageKite what IP rate limits apply
        'ips_per_sec-secs': 900} # ... to this tunnel's incoming traffic.



def ZkAuth(domain):
    # This usually does nothing, adjust CAPABILITIES above if you want to
    # take this code path.
    #
    # In reality this method would both decode the incoming "auth domain"
    # string and verify the signed challenge against a shared secret, and
    # then reformat the returned quota values as a dynamic DNS response...
    #
    return {'hostname': domain, 'alias': '', 'ips': ['0.0.255.255']}


def P(string):
    # Delete the sys.stderr line if you're not debugging.
    sys.stderr.write('>> ' + string + '\n')
    print(string)


def ProcessArgs(args, server=False):
    o, a = getopt.getopt(args, 'a:z:',
        ([] if server else ['capabilities', 'server']) +
        ['auth=', 'zk-auth='])

    for opt, arg in o:
        sys.stderr.write('<< %s=%s\n' % (opt, arg))

        if opt == '--capabilities':
            P(CAPABILITIES)
            return

        if opt == '--server':
            ServerLoop()
            return

        if opt in ('-z', '--zk-auth'):
            P(json.dumps(ZkAuth(arg), indent=None))
            return

        if opt in ('-a', '--auth'):
            P(json.dumps(Auth(arg), indent=None))
            return


def ServerLoop():
    while True:
        line = sys.stdin.readline()
        if not line:
            return

        args = line.strip().split()
        if args and not args[0][:2] == '--':
            args[0] = '--' + args[0]
        ProcessArgs(args, server=True)


if __name__ == '__main__':
    ProcessArgs(sys.argv[1:])



"""
This is a "dropper template".  A dropper is a single-purpose PageKite
back-end connector which embeds its own configuration.
"""
##############################################################################

from __future__ import absolute_import

LICENSE = """\
This file is part of pagekite.py.
Copyright 2010-2020, the Beanstalks Project ehf. and Bjarni Runar Einarsson

This program is free software: you can redistribute it and/or modify it under
the terms of the  GNU  Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful,  but  WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see: <http://www.gnu.org/licenses/>
"""
##############################################################################
import sys
import pagekite.pk as pk
import pagekite.httpd as httpd

if __name__ == "__main__":
  kn = '@KITENAME@'
  ss = '@SECRET@'
  if len(sys.argv) == 1:
    sys.argv.extend([
      '--daemonize',
      '--runas=nobody',
      '--logfile=/tmp/pagekite-%s.log' % kn,
    ])
  sys.argv[1:1] = [
    '--clean',
    '--noloop',
    '--nocrashreport',
    '--defaults',
    '--kitename=%s' % kn,
    '--kitesecret=%s' % ss,
    '--all'
  ]
  sys.argv.extend('@ARGS@'.split())

  pk.Main(pk.PageKite, pk.Configure,
          http_handler=httpd.UiRequestHandler,
          http_server=httpd.UiHttpServer)




"""
Logging.
"""

from __future__ import absolute_import

##############################################################################
LICENSE = """\
This file is part of pagekite.py.
Copyright 2010-2020, the Beanstalks Project ehf. and Bjarni Runar Einarsson

This program is free software: you can redistribute it and/or modify it under
the terms of the  GNU  Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful,  but  WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see: <http://www.gnu.org/licenses/>
"""
##############################################################################
import threading
import time
import sys

from . import compat, common
from .compat import *
from .common import *

syslog = compat.syslog
org_stdout = sys.stdout

DEBUG_IO = False

LOG = []
LOG_LINE = 0
LOG_LENGTH = 300
LOG_THRESHOLD = 256 * 1024
LOG_LOCK = threading.Lock()

LOG_LEVEL_NONE = 1
LOG_LEVEL_ERR = 2
LOG_LEVEL_WARN = 3
LOG_LEVEL_INFO = 4
LOG_LEVEL_MACH = 5
LOG_LEVEL_DEBUG = 6
LOG_LEVEL_DEFAULT = LOG_LEVEL_INFO
LOG_LEVELS = {
  'none': LOG_LEVEL_NONE,
  'err': LOG_LEVEL_ERR,
  'errors': LOG_LEVEL_ERR,
  'warn': LOG_LEVEL_WARN,
  'warnings': LOG_LEVEL_WARN,
  'info': LOG_LEVEL_INFO,
  'mach': LOG_LEVEL_MACH,
  'machine': LOG_LEVEL_MACH,
  'debug': LOG_LEVEL_DEBUG,
  'full': LOG_LEVEL_DEBUG,
  'all': LOG_LEVEL_DEBUG,
  0: 'none',
  LOG_LEVEL_NONE: 'none',
  LOG_LEVEL_ERR: 'err',
  LOG_LEVEL_WARN: 'warn',
  LOG_LEVEL_INFO: 'info',
  LOG_LEVEL_MACH: 'mach',
  LOG_LEVEL_DEBUG: 'debug'}
LOG_LEVEL_DEFNAME = LOG_LEVELS[LOG_LEVEL_DEFAULT]

LOG_LEVEL = LOG_LEVEL_DEFAULT


def LogValues(values, testtime=None):
  global LOG, LOG_LINE, LOG_LAST_TIME
  now = int(testtime or time.time())
  words = [('ts', '%x' % now),
           ('t',  '%s' % ts_to_iso(now)),
           ('ll', '%x' % LOG_LINE)]
  words.extend([(kv[0], ('%s' % kv[1]).replace('\t', ' ')
                                      .replace('\r', ' ')
                                      .replace('\n', ' ')
                                      .replace('; ', ', ')
                                      .strip()) for kv in values])
  wdict = dict(words)
  LOG_LINE += 1
  LOG.append(wdict)
  while len(LOG) > LOG_LENGTH:
    LOG[0:(LOG_LENGTH//10)] = []

  return (words, wdict)

def LogSyslog(values, wdict=None, words=None, level=LOG_LEVEL_INFO):
  global LOG_LEVEL
  if level > LOG_LEVEL: return
  if values:
    words, wdict = LogValues(values)
  if level <= LOG_LEVEL_ERR or ('err' in wdict):
    syslog.syslog(syslog.LOG_ERR, '; '.join(['='.join(x) for x in words]))
  elif level <= LOG_LEVEL_INFO:
    syslog.syslog(syslog.LOG_INFO, '; '.join(['='.join(x) for x in words]))
  else:
    syslog.syslog(syslog.LOG_DEBUG, '; '.join(['='.join(x) for x in words]))

def LogToFile(values, wdict=None, words=None, level=LOG_LEVEL_INFO):
  global LOG_LEVEL
  if level > LOG_LEVEL: return
  if values:
    words, wdict = LogValues(values)
  try:
    global LogFile
    with LOG_LOCK:
      LogFile.write('; '.join(['='.join(x) for x in words]))
      LogFile.write('\n')
      LogFile.flush()
  except (OSError, IOError):
    # Avoid crashing if the disk fills up or something lame like that
    pass

def LogToMemory(values, wdict=None, words=None, level=LOG_LEVEL_INFO):
  global LOG_LEVEL
  if values and (level <= LOG_LEVEL):
    with LOG_LOCK:
      LogValues(values)

def FlushLogMemory():
  global LOG
  for l in LOG:
    Log(None, wdict=l, words=[(w, l[w]) for w in l], level=LOG_LEVEL)

def LogError(msg, parms=None):
  emsg = [('err', msg)]
  if parms: emsg.extend(parms)
  Log(emsg, level=LOG_LEVEL_ERR)

  if common.gYamon:
    common.gYamon.vadd('errors', 1, wrap=1000000)

def LogWarning(msg, parms=None):
  emsg = [('warn', msg)]
  if parms: emsg.extend(parms)
  Log(emsg, level=LOG_LEVEL_WARN)

def LogDebug(msg, parms=None):
  emsg = [('debug', msg)]
  if parms: emsg.extend(parms)
  Log(emsg, level=LOG_LEVEL_DEBUG)

def LogInfo(msg, parms=None):
  emsg = [('info', msg)]
  if parms: emsg.extend(parms)
  Log(emsg, level=LOG_LEVEL_INFO)

def ResetLog():
  global LogFile, Log, org_stdout
  LogFile = org_stdout
  Log = LogToMemory

ResetLog()




"""
Selectables are low level base classes which cooperate with our select-loop.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

##############################################################################
LICENSE = """\
This file is part of pagekite.py.
Copyright 2010-2020, the Beanstalks Project ehf. and Bjarni Runar Einarsson

This program is free software: you can redistribute it and/or modify it under
the terms of the  GNU  Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful,  but  WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see: <http://www.gnu.org/licenses/>
"""
##############################################################################

import errno
import os
import re
import struct
import threading
import time
import zlib

from pagekite.compat import *
from pagekite.common import *
from pagekite.proto.proto import HTTP_Unavailable
from pagekite.proto.ws_abnf import ABNF
import pagekite.logging as logging
import pagekite.compat as compat
import pagekite.common as common


def obfuIp(ip):
  quads = ('%s' % ip).replace(':', '.').split('.')
  return '~%s' % '.'.join([q for q in quads[-2:]])


SELECTABLE_LOCK = threading.RLock()  # threading.Lock() will deadlock on pypy!
SELECTABLE_ID = 0
SELECTABLES = set([])
def getSelectableId(what):
  global SELECTABLES, SELECTABLE_ID, SELECTABLE_LOCK
  with SELECTABLE_LOCK:
    count = 0
    SELECTABLE_ID += 1
    SELECTABLE_ID %= 0x20000
    while SELECTABLE_ID in SELECTABLES:
      SELECTABLE_ID += 1
      SELECTABLE_ID %= 0x20000
      count += 1
      if count > 0x20000:
        raise ValueError('Too many conns!')
    SELECTABLES.add(SELECTABLE_ID)
    return SELECTABLE_ID


class Selectable(object):
  """A wrapper around a socket, for use with select."""

  HARMLESS_ERRNOS = (errno.EINTR, errno.EAGAIN, errno.ENOMEM, errno.EBUSY,
                     errno.EDEADLK, errno.EWOULDBLOCK, errno.ENOBUFS,
                     errno.EALREADY)

  def __init__(self, fd=None, address=None, on_port=None, maxread=None,
                     ui=None, tracked=True, bind=None, backlog=100):
    self.fd = None

    try:
      if bind and bind[0] and re.match(r'^\d+\.\d+\.\d+\.\d+$', bind[0]):
        raise ValueError('Avoid INET6 for IPv4 hosts')
      self.SetFD(fd or rawsocket(socket.AF_INET6, socket.SOCK_STREAM), six=True)
      if bind:
        self.fd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.fd.bind(bind)
        self.fd.listen(backlog)
        self.fd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    except:
      self.SetFD(fd or rawsocket(socket.AF_INET, socket.SOCK_STREAM))
      if bind:
        self.fd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.fd.bind(bind)
        self.fd.listen(backlog)
        self.fd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    self.address = address
    self.on_port = on_port
    self.created = self.bytes_logged = time.time()
    self.lock = threading.RLock()
    self.last_activity = 0
    self.dead = False
    self.ui = ui

    # Quota-related stuff
    self.quota = None
    self.q_conns = None
    self.q_days = None

    # Read-related variables
    self.maxread = maxread or common.MAX_READ_BYTES
    self.read_bytes = self.all_in = 0
    self.read_eof = False
    self.peeking = False
    self.peeked = 0
    self.retry_delays = [0.0, 0.02, 0.05]

    # Write-related variables
    self.wrote_bytes = self.all_out = 0
    self.write_blocked = ''
    self.write_speed = 102400
    self.write_eof = False

    # Flow control v2
    self.acked_kb_delta = 0

    # Compression stuff
    self.zw = None
    self.zlevel = 1
    self.zreset = False

    # This is the default, until we switch to Websockets...
    self.use_websocket = False
    self.ws_zero_mask = False

    # Logging
    self.sstate = 'new'
    self.alt_id = None
    self.countas = 'selectables_live'
    self.sid = self.gsid = getSelectableId(self.countas)

    if address:
      addr = address or ('x.x.x.x', 'x')
      self.log_id = 's%x/%s:%s' % (self.sid, obfuIp(addr[0]), addr[1])
    else:
      self.log_id = 's%x' % self.sid

    if common.gYamon:
      common.gYamon.vadd(self.countas, 1)
      common.gYamon.vadd('selectables', 1)

  def ExtendSSLRetryDelays(self):
    self.LogDebug('Extended SSL Write retries on %s' % self)
    self.retry_delays = [0.0, 0.02, 0.05, 0.30, 0.70, 1.5, 2.0]

  def CountAs(self, what):
    with self.lock:
      if common.gYamon:
        common.gYamon.vadd(self.countas, -1)
        common.gYamon.vadd(what, 1)
      self.countas = what

  def Cleanup(self, close=True):
    self.peeked = self.zw = ''
    self.Die(discard_buffer=True)
    if close:
      if self.fd:
        if logging.DEBUG_IO:
          self.LogDebug('Closing FD: %s' % self)
        self.fd.close()
    self.fd = None
    if not self.dead:
      self.dead = True
      self.sstate = 'dead'
      self.CountAs('selectables_dead')
      if close:
        self.LogTraffic(final=True)
    try:
      global SELECTABLES, SELECTABLE_LOCK
      with SELECTABLE_LOCK:
        SELECTABLES.remove(self.gsid)
    except KeyError:
      pass

  def __del__(self):
    # Important: This can run at random times, especially under pypy, so all
    #            locks must be re-entrant (RLock), otherwise we deadlock.
    try:
      with self.lock:
        if common.gYamon and self.countas:
          common.gYamon.vadd(self.countas, -1)
          common.gYamon.vadd('selectables', -1)
          self.countas = None
    except AttributeError:
      pass

  def __str__(self):
    return '%s: %s<%s|%s%s%s>' % (self.log_id, self.__class__, self.sstate,
                                  self.read_eof and '-' or 'r',
                                  self.write_eof and '-' or 'w',
                                  len(self.write_blocked))

  def __html__(self):
    try:
      peer = self.fd.getpeername()
    except:
      peer = ('x.x.x.x', 'x')

    try:
      sock = self.fd.getsockname()
    except:
      sock = ('x.x.x.x', 'x')

    return ('<b>Outgoing ZChunks</b>: %s<br>'
            '<b>Buffered bytes</b>: %s<br>'
            '<b>Remote address</b>: %s<br>'
            '<b>Local address</b>: %s<br>'
            '<b>Bytes in / out</b>: %s / %s<br>'
            '<b>Created</b>: %s<br>'
            '<b>Status</b>: %s<br>'
            '\n') % (self.zw and ('level %d' % self.zlevel) or 'off',
                     len(self.write_blocked),
                     self.dead and '-' or (obfuIp(peer[0]), peer[1]),
                     self.dead and '-' or (obfuIp(sock[0]), sock[1]),
                     self.all_in + self.read_bytes,
                     self.all_out + self.wrote_bytes,
                     time.strftime('%Y-%m-%d %H:%M:%S',
                                   time.localtime(self.created)),
                     self.sstate)

  def ResetZChunks(self):
    with self.lock:
      if self.zw:
        self.zreset = True
        self.zw = zlib.compressobj(self.zlevel)

  def EnableZChunks(self, level=1):
    with self.lock:
      self.zlevel = level
      self.zw = zlib.compressobj(level)

  def EnableWebsockets(self):
    with self.lock:
      self.use_websocket = True
      self.ws_zero_mask = (
        hasattr(self.fd, 'get_cipher_name') or hasattr(self.fd, 'getpeercert'))

  def SetFD(self, fd, six=False):
    if self.fd:
      self.fd.close()
    self.fd = fd
    if fd:
      self.fd.setblocking(0)
    try:
      if six:
        self.fd.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
      # This hurts mobile devices, let's try living without it
      #self.fd.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
      #self.fd.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 60)
      #self.fd.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 10)
      #self.fd.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 1)
    except:
      pass

  def SetConn(self, conn):
    self.SetFD(conn.fd)
    self.log_id = conn.log_id
    self.read_bytes = conn.read_bytes
    self.wrote_bytes = conn.wrote_bytes

  def Log(self, values, level=logging.LOG_LEVEL_DEFAULT):
    if self.log_id: values.append(('id', self.log_id))
    logging.Log(values, level=level)

  def LogError(self, error, params=None):
    values = params or []
    if self.log_id: values.extend([('id', self.log_id), ('s', self.sstate)])
    logging.LogError(error, values)

  def LogDebug(self, message, params=None):
    values = params or []
    if self.log_id: values.extend([('id', self.log_id), ('s', self.sstate)])
    logging.LogDebug(message, values)

  def LogWarning(self, warning, params=None):
    values = params or []
    if self.log_id: values.append(('id', self.log_id))
    logging.LogWarning(warning, values)

  def LogInfo(self, message, params=None):
    values = params or []
    if self.log_id: values.extend([('id', self.log_id), ('s', self.sstate)])
    logging.LogInfo(message, values)

  def LogTrafficStatus(self, final=False):
    if self.ui:
      self.ui.Status('traffic')

  def LogTraffic(self, final=False):
    if self.wrote_bytes or self.read_bytes:
      now = time.time()
      self.all_out += self.wrote_bytes
      self.all_in += self.read_bytes

      self.LogTrafficStatus(final)

      if common.gYamon:
        common.gYamon.vadd("bytes_all", self.wrote_bytes
                                        + self.read_bytes, wrap=1000000000)

      log_info = [('wrote', '%d' % self.wrote_bytes),
                  ('wbps', '%d' % self.write_speed),
                  ('read', '%d' % self.read_bytes)]
      if self.acked_kb_delta:
        log_info.append(('delta', '%d' % self.acked_kb_delta))
      if final:
        log_info.append(('eof', '1'))
      self.Log(log_info)

      self.bytes_logged = now
      self.wrote_bytes = self.read_bytes = 0
    elif final:
      self.Log([('eof', '1')], level=logging.LOG_LEVEL_MACH)

  def SayHello(self):
    pass

  def ProcessData(self, data):
    self.LogError('Selectable::ProcessData: Should be overridden!')
    return False

  def ProcessEof(self):
    if self.read_eof and self.write_eof and not self.write_blocked:
      self.Cleanup()
      return False
    return True

  def ProcessEofRead(self):
    self.read_eof = True
    self.LogError('Selectable::ProcessEofRead: Should be overridden!')
    return False

  def ProcessEofWrite(self):
    self.write_eof = True
    self.LogError('Selectable::ProcessEofWrite: Should be overridden!')
    return False

  def EatPeeked(self, eat_bytes=None, keep_peeking=False):
    if not self.peeking: return
    if eat_bytes is None: eat_bytes = self.peeked
    discard = ''
    while len(discard) < eat_bytes:
      try:
        bytecount = eat_bytes - len(discard)
        self.sstate = 'eat(%d)' % bytecount
        discard += self.fd.recv(bytecount)
      except socket.error as err:
        (errno, msg) = err.args
        self.LogInfo('Error reading (%d/%d) socket: %s (errno=%s)' % (
                       eat_bytes, self.peeked, msg, errno))
        time.sleep(0.1)

    if logging.DEBUG_IO:
      print('===[ ATE %d PEEKED BYTES ]===\n' % eat_bytes)
    self.sstate = 'ate(%d)' % eat_bytes
    self.peeked -= eat_bytes
    self.peeking = keep_peeking
    return

  def ReadData(self, maxread=None):
    if self.read_eof:
      return False

    now = time.time()
    maxread = maxread or self.maxread
    try:
      if self.peeking:
        self.sstate = 'peek(%d)' % maxread
        data = s(self.fd.recv(maxread, socket.MSG_PEEK))
        self.peeked = len(data)
        if logging.DEBUG_IO:
          print('<== PEEK =[%s]==(\n%s)==' % (self, data[:320]))
      else:
        self.sstate = 'read(%d)' % maxread
        data = s(self.fd.recv(maxread))
        if logging.DEBUG_IO:
          print('<== IN =[%s]==(\n%s)==' % (self, data[:160]))
      self.sstate = 'data(%d)' % len(data)
    except (SSL.WantReadError, SSL.WantWriteError):
      self.sstate += '/SSL.WRE'
      return True
    except IOError as err:
      self.sstate += '/ioerr=%s' % (err.errno,)
      if err.errno not in self.HARMLESS_ERRNOS:
        self.LogDebug('Error reading socket: %s (%s)' % (err, err.errno))
        return False
      else:
        return True
    except (SSL.Error, SSL.ZeroReturnError, SSL.SysCallError) as err:
      self.sstate += '/SSL.Error'
      self.LogDebug('Error reading socket (SSL): %s' % err)
      return False
    except socket.error as err:
      (errno, msg) = err.args
      self.sstate += '/sockerr=%s' % (err.errno,)
      if errno in self.HARMLESS_ERRNOS:
        return True
      else:
        self.LogInfo('Error reading socket: %s (errno=%s)' % (msg, errno))
        return False

    try:
      self.last_activity = now
      if data is None or data == '':
        self.sstate += '/EOF'
        self.read_eof = True
        if logging.DEBUG_IO:
          print('<== IN =[%s]==(EOF)==' % self)
        return self.ProcessData('')
      else:
        if not self.peeking:
          self.read_bytes += len(data)
          if self.acked_kb_delta:
            self.acked_kb_delta += (len(data)/1024)
          if self.read_bytes > logging.LOG_THRESHOLD: self.LogTraffic()
        return self.ProcessData(data)
    finally:
      self.sstate = (self.dead and 'dead' or 'idle')

  def RecordProgress(self, skb):
    if skb >= 0:
      all_read = (self.all_in + self.read_bytes) // 1024
      self.acked_kb_delta = max(1, all_read - skb)

  def Send(self, data, try_flush=False, activity=False,
                       just_buffer=False, allow_blocking=False):
    global SEND_MAX_BYTES
    self.write_speed = int((self.wrote_bytes + self.all_out)
                           / max(1, (time.time() - self.created)))  # Integer division

    # If we're already blocked, just buffer unless explicitly asked to flush.
    if ((just_buffer) or
        ((not try_flush) and
         (len(self.write_blocked) > 0 or compat.SEND_ALWAYS_BUFFERS))):
      self.write_blocked += str(''.join(data))
      return True

    pending = ''.join([self.write_blocked, str(''.join(data))])
    self.write_blocked = ''
    if pending:
      try:
        sent = None
        send_bytes = min(len(pending), SEND_MAX_BYTES)
        send_buffer = b(pending[:send_bytes])
        self.sstate = 'send(%d)' % (send_bytes)
        for try_wait in self.retry_delays:
          try:
            sent = self.fd.send(send_buffer)
            if logging.DEBUG_IO:
              print(('==> OUT =[%s: %d/%d bytes]==(\n%s)=='
                     ) % (self, sent, send_bytes, send_buffer[:min(320, sent)]))
            self.wrote_bytes += sent
            break
          except (SSL.WantWriteError, SSL.WantReadError) as err:
            SEND_MAX_BYTES = min(4096, SEND_MAX_BYTES)  # Maybe this will help?
            if logging.DEBUG_IO:
              print('=== WRITE SSL RETRY: =[%s: %s bytes]==' % (self, send_bytes))
            self.sstate = 'send/SSL.WRE(%d,%.1f)' % (send_bytes, try_wait)
            time.sleep(try_wait)
        if sent is None:
          self.sstate += '/retries'
          self.LogInfo(
              'Error sending: Too many SSL write retries (SEND_MAX_BYTES=%d)'
              % SEND_MAX_BYTES)
          self.ProcessEofWrite()
          return False
      except IOError as err:
        self.sstate += '/ioerr=%s' % (err.errno,)
        if err.errno not in self.HARMLESS_ERRNOS:
          self.LogInfo('Error sending: %s' % err)
          self.ProcessEofWrite()
          return False
        else:
          if logging.DEBUG_IO:
            print('=== WRITE HICCUP: =[%s: %s bytes]==' % (self, send_bytes))
      except socket.error as err:
        (errno, msg) = err.args
        self.sstate += '/sockerr=%s' % (errno,)
        if errno not in self.HARMLESS_ERRNOS:
          self.LogInfo('Error sending: %s (errno=%s)' % (msg, errno))
          self.ProcessEofWrite()
          return False
        else:
          if logging.DEBUG_IO:
            print('=== WRITE HICCUP: =[%s: %s bytes]==' % (self, send_bytes))
      except (SSL.Error, SSL.ZeroReturnError, SSL.SysCallError) as err:
        self.sstate += '/SSL.Error'
        self.LogInfo(
            'Error sending (SSL, SEND_MAX_BYTES=%d): %s'
            % (SEND_MAX_BYTES, err))
        SEND_MAX_BYTES = min(4096, SEND_MAX_BYTES)  # Maybe this will help?
        self.ProcessEofWrite()
        return False
      except AttributeError:
        self.sstate += '/AttrError'
        # This has been seen in the wild, is most likely some sort of
        # race during shutdown. :-(
        self.LogInfo('AttributeError, self.fd=%s' % self.fd)
        self.ProcessEofWrite()
        return False

      self.write_blocked = pending[sent:]

    if activity:
      self.last_activity = time.time()

    if self.wrote_bytes >= logging.LOG_THRESHOLD:
      self.LogTraffic()

    if self.write_eof and not self.write_blocked:
      self.ProcessEofWrite()

    self.sstate = (self.dead and 'dead' or 'idle')
    return True

  def SendChunked(self, data, compress=True, zhistory=None, just_buffer=False):
    if self.use_websocket:
      with self.lock:
        return self.Send(
          [ABNF.create_frame(
            ''.join(data), ABNF.OPCODE_BINARY, 1, self.ws_zero_mask).format()],
          activity=False, try_flush=(not just_buffer), just_buffer=just_buffer)

    rst = ''
    if self.zreset:
      self.zreset = False
      rst = 'R'

    # Stop compressing streams that just get bigger.
    if zhistory and (zhistory[0] < zhistory[1]): compress = False
    with self.lock:
      try:
        sdata = ''.join(data)
        if self.zw and compress and len(sdata) > 64:
          try:
            zdata = s(self.zw.compress(b(sdata)) + self.zw.flush(zlib.Z_SYNC_FLUSH))
            if zhistory:
              zhistory[0] = len(sdata)
              zhistory[1] = len(zdata)
            return self.Send(['%xZ%x%s\r\n' % (len(sdata), len(zdata), rst), zdata],
                             activity=False,
                             try_flush=(not just_buffer), just_buffer=just_buffer)
          except zlib.error:
            logging.LogError('Error compressing, resetting ZChunks.')
            self.ResetZChunks()

        return self.Send(['%x%s\r\n' % (len(sdata), rst), sdata],
                         activity=False,
                         try_flush=(not just_buffer), just_buffer=just_buffer)
      except UnicodeDecodeError:
        logging.LogError('UnicodeDecodeError in SendChunked, wtf?')
        return False

  def Flush(self, loops=50, wait=False, allow_blocking=False):
    while (loops != 0 and
           len(self.write_blocked) > 0 and
           self.Send([], try_flush=True, activity=False,
                         allow_blocking=allow_blocking)):
      if wait and len(self.write_blocked) > 0:
        time.sleep(0.1)
      logging.LogDebug('Flushing...')
      loops -= 1

    if self.write_blocked: return False
    return True

  def IsReadable(s, now):
    return (s.fd and (not s.read_eof)
                 and (s.acked_kb_delta < (3 * s.maxread/1024)))

  def IsBlocked(s):
    return (s.fd and (len(s.write_blocked) > 0))

  def IsDead(s):
    return (s.read_eof and s.write_eof and not s.write_blocked)

  def Die(self, discard_buffer=False):
    if discard_buffer:
      self.write_blocked = ''
    self.read_eof = self.write_eof = True
    return True

  def HTTP_Unavail(self, config, where, proto, host, **kwargs):
    kwargs['frame_url'] = config.error_url
    if self.fd and where in ('FE', 'fe'):
      try:
        kwargs['relay_sockname'] = self.fd.getsockname()
      except:
        kwargs['relay_sockname'] = None

    # Do we have a more specific error URL for this domain? This is a
    # white-label feature, for folks not wanting to hit the PageKite.net
    # servers at all. In case of a match, we also disable mention of
    # PageKite itself in the HTML boilerplate.
    dparts = host.split(':')[0].split('.')
    while dparts:
      fu = config.error_urls.get('.'.join(dparts), None)
      if fu is not None:
        kwargs['frame_url'] = fu
        kwargs['advertise'] = False
        break
      dparts.pop(0)

    return HTTP_Unavailable(where, proto, host, **kwargs)


class LineParser(Selectable):
  """A Selectable which parses the input as lines of text."""

  def __init__(self, fd=None, address=None, on_port=None,
                     ui=None, tracked=True):
    Selectable.__init__(self, fd, address, on_port, ui=ui, tracked=tracked)
    self.leftovers = ''

  def __html__(self):
    return Selectable.__html__(self)

  def Cleanup(self, close=True):
    Selectable.Cleanup(self, close=close)
    self.leftovers = ''

  def ProcessData(self, data):
    lines = (self.leftovers+data).splitlines(True)
    self.leftovers = ''

    while lines:
      line = lines.pop(0)
      if line.endswith('\n'):
        if self.ProcessLine(line, lines) is False:
          return False
      else:
        if not self.peeking: self.leftovers += line

    if self.read_eof: return self.ProcessEofRead()
    return True

  def ProcessLine(self, line, lines):
    self.LogError('LineParser::ProcessLine: Should be overridden!')
    return False


TLS_CLIENTHELLO = '%c' % 0o26
SSL_CLIENTHELLO = '\x80'
XML_PREAMBLE = '<?xml'
XMPP_REGEXP = re.compile("<[^>]+\sto=([^\s>]+)[^>]*>")

class MagicProtocolParser(LineParser):
  """A Selectable which recognizes HTTP, TLS or XMPP preambles."""

  def __init__(self, fd=None, address=None, on_port=None, ui=None):
    LineParser.__init__(self, fd, address, on_port, ui=ui, tracked=False)
    self.leftovers = ''
    self.might_be_tls = True
    self.is_tls = False
    self.my_tls = False

  def __html__(self):
    return ('<b>Detected TLS</b>: %s<br>'
            '%s') % (self.is_tls, LineParser.__html__(self))

  def ProcessData(self, data):
    # Uncomment when adding support for new protocols:
    #
    #print(('DATA: >%s<'
    #       ) % ' '.join(['%2.2x' % ord(d) for d in data]))

    if self.might_be_tls:
      self.might_be_tls = False
      if not (data.startswith(TLS_CLIENTHELLO) or
              data.startswith(SSL_CLIENTHELLO)):
        self.EatPeeked()

        # Only works if the full request is in the first data packet.
        if data.startswith(XML_PREAMBLE):
          server = self.GetXMPPServer(data)
          if server:
            return self.ProcessProto(data, 'xmpp', server)

        return LineParser.ProcessData(self, data)

      self.is_tls = True

    if self.is_tls:
      return self.ProcessTls(data)
    else:
      self.EatPeeked()
      return LineParser.ProcessData(self, data)

  def GetMsg(self, data):
    mtype, ml24, mlen = struct.unpack('>BBH', data[0:4])
    mlen += ml24 * 0x10000
    return mtype, data[4:4+mlen], data[4+mlen:]

  def GetClientHelloExtensions(self, msg):
    # Ugh, so many magic numbers! These are accumulated sizes of
    # the different fields we are ignoring in the TLS headers.
    slen = struct.unpack('>B', msg[34])[0]
    cslen = struct.unpack('>H', msg[35+slen:37+slen])[0]
    cmlen = struct.unpack('>B', msg[37+slen+cslen])[0]
    extofs = 34+1+2+1+2+slen+cslen+cmlen
    if extofs < len(msg): return msg[extofs:]
    return None

  def GetSniNames(self, extensions):
    names = []
    while extensions:
      etype, elen = struct.unpack('>HH', extensions[0:4])
      if etype == 0:
        # OK, we found an SNI extension, get the list.
        namelist = extensions[6:4+elen]
        while namelist:
          ntype, nlen = struct.unpack('>BH', namelist[0:3])
          if ntype == 0: names.append(namelist[3:3+nlen].lower())
          namelist = namelist[3+nlen:]
      extensions = extensions[4+elen:]
    return names

  def GetSni(self, data):
    hello, vmajor, vminor, mlen = struct.unpack('>BBBH', data[0:5])
    data = data[5:]
    sni = []
    while data:
      mtype, msg, data = self.GetMsg(data)
      if mtype == 1:
        # ClientHello!
        sni.extend(self.GetSniNames(self.GetClientHelloExtensions(msg)))
    return sni

  def GetXMPPServer(self, data):
    match = XMPP_REGEXP.search(data)
    if match and match.group(1):
      server = match.group(1)
      if server[:1] in ('"', "'"):
        server = server[1:-1]
      if '@' in server:
        server = server.split('@')[-1]
      return server
    else:
      return None

  def ProcessTls(self, data, domain=None):
    self.LogError('MagicProtocolParser::ProcessTls: Should be overridden!')
    return False

  def ProcessProto(self, data, proto, domain):
    self.LogError('MagicProtocolParser::ProcessProto: Should be overridden!')
    return False


class ChunkParser(Selectable):
  """A Selectable which parses the input as chunks."""

  def __init__(self, fd=None, address=None, on_port=None, ui=None):
    Selectable.__init__(self, fd, address, on_port, ui=ui)
    self.want_cbytes = 0
    self.want_bytes = 0
    self.compressed = False
    self.header = ''
    self.chunk = ''
    self.zr = None
    self.websocket_key = None

  def ProcessData(self, *args, **kwargs):
    if self.use_websocket:
      return self.ProcessWebsocketData(*args, **kwargs)
    else:
      return self.ProcessPageKiteData(*args, **kwargs)

  def __html__(self):
    return Selectable.__html__(self)

  def Cleanup(self, close=True):
    Selectable.Cleanup(self, close=close)
    self.zr = self.chunk = self.header = None

  def PrepareWebsockets(self):
    self.websocket_key = os.urandom(16)

  def ProcessWebsocketData(self, data):
    loops = 150
    happy = more = True
    while happy and more and (loops > 0):
      loops -= 1

      if self.peeking:
        self.header = ''
        self.chunk = ''

      self.header += (data or '')
      try:
        ws_frame, data = ABNF.parse(self.header)
        more = data and (len(data) > 0)
      except ValueError as err:
        self.LogError('ChunkParser::ProcessData: %s' % err)
        self.Log([('bad_data', data)])
        return False

      if ws_frame and ws_frame.length == len(ws_frame.data):
        # We have a complete frame, process it!
        self.header = ''
        happy = self.ProcessChunk(ws_frame.data) if ws_frame.data else True
      else:
        if self.read_eof:
          return self.ProcessEofRead()
        # Frame is incomplete, but there were no errors: we're done for now.
        return True

    if happy and more:
      self.LogError('Unprocessed data: %s' % data)
      raise BugFoundError('Too much data')
    elif self.read_eof:
      return self.ProcessEofRead() and happy
    else:
      return happy

    return False  # Not reached

  def ProcessPageKiteData(self, data):
    loops = 150
    result = more = True
    while result and more and (loops > 0):
      loops -= 1

      if self.peeking:
        self.want_cbytes = 0
        self.want_bytes = 0
        self.header = ''
        self.chunk = ''

      if self.want_bytes == 0:
        self.header += (data or '')
        if self.header.find('\r\n') < 0:
          if self.read_eof:
            return self.ProcessEofRead()
          return True

        try:
          size, data = self.header.split('\r\n', 1)
          self.header = ''

          if size.endswith('R'):
            self.zr = zlib.decompressobj()
            size = size[0:-1]

          if 'Z' in size:
            csize, zsize = size.split('Z')
            self.compressed = True
            self.want_cbytes = int(csize, 16)
            self.want_bytes = int(zsize, 16)
          else:
            self.compressed = False
            self.want_bytes = int(size, 16)

        except ValueError as err:
          self.LogError('ChunkParser::ProcessData: %s' % err)
          self.Log([('bad_data', data)])
          return False

        if self.want_bytes == 0:
          return False

      process = data[:self.want_bytes]
      data = more = data[self.want_bytes:]

      self.chunk += process
      self.want_bytes -= len(process)

      if self.want_bytes == 0:
        if self.compressed:
          try:
            if not self.zr:
              self.zr = zlib.decompressobj()
            cchunk = s(self.zr.decompress(b(self.chunk)))
          except zlib.error:
            cchunk = ''

          if len(cchunk) != self.want_cbytes:
            result = self.ProcessCorruptChunk(self.chunk)
          else:
            result = self.ProcessChunk(cchunk)
        else:
          result = self.ProcessChunk(self.chunk)
        self.chunk = ''

    if result and more:
      self.LogError('Unprocessed data: %s' % data)
      raise BugFoundError('Too much data')
    elif self.read_eof:
      return self.ProcessEofRead() and result
    else:
      return result

  def ProcessCorruptChunk(self, chunk):
    self.LogError('ChunkParser::ProcessData: ProcessCorruptChunk not overridden!')
    return False

  def ProcessChunk(self, chunk):
    self.LogError('ChunkParser::ProcessData: ProcessChunk not overridden!')
    return False



##############################################################################
LICENSE = """\
This file is part of pagekite.py.
Copyright 2010-2020, the Beanstalks Project ehf. and Bjarni Runar Einarsson

This program is free software: you can redistribute it and/or modify it under
the terms of the  GNU  Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful,  but  WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see: <http://www.gnu.org/licenses/>
"""
##############################################################################



#!/usr/bin/env python
"""
The program manual!
"""

from __future__ import absolute_import
from __future__ import print_function

import os
import re
import time

from .common import *
from .compat import ts_to_iso

MAN_NAME = ("""\
    pagekite.py - Make localhost servers publicly visible
""")
MAN_SYNOPSIS = ("""\
    <b>pagekite.py</b> [<a>--options</a>] [<a>service</a>] <a>kite-name</a> [<a>+flags</a>]
""")
MAN_DESCRIPTION = ("""\
    PageKite is a system for exposing <tt>localhost</tt> servers to the
    public Internet.  It is most commonly used to make local web servers or
    SSH servers publicly visible, although almost any TCP-based protocol can
    work if the client knows how to use an HTTP proxy.

    PageKite uses a combination of tunnels and reverse proxies to compensate
    for the fact that <tt>localhost</tt> usually does not have a public IP
    address and is often subject to adverse network conditions, including
    aggressive firewalls and multiple layers of NAT.

    This program implements both ends of the tunnel: the local "back-end"
    and the remote "front-end" reverse-proxy relay.  For convenience,
    <b>pagekite.py</b> also includes a basic HTTP server for quickly exposing
    files and directories to the World Wide Web for casual sharing and
    collaboration.
""")
MAN_EXAMPLES = ("""\
    <pre>Basic usage, gives <tt>http://localhost:80/</tt> a public name:
    $ pagekite.py NAME.pagekite.me

    To expose specific folders, files or use alternate local ports:
    $ pagekite.py /a/path/ NAME.pagekite.me +indexes  # built-in HTTPD
    $ pagekite.py *.html   NAME.pagekite.me           # built-in HTTPD
    $ pagekite.py 3000     NAME.pagekite.me           # HTTPD on 3000

    To expose multiple local servers (SSH and HTTP):
    $ pagekite.py ssh://NAME.pagekite.me AND 3000 NAME.pagekite.me</pre>
""")
MAN_KITES = ("""\
    The most comman usage of <b>pagekite.py</b> is as a back-end, where it
    is used to expose local services to the outside world.

    Examples of services are: a local HTTP server, a local SSH server,
    a folder or a file.

    A service is exposed by describing it on the command line, along with the
    desired public kite name. If a kite name is requested which does not already
    exist in the configuration file and program is run interactively, the user
    will be prompted and given the option of signing up and/or creating a new
    kite using the <b>pagekite.net</b> service.

    Multiple services and kites can be specified on a single command-line,
    separated by the word 'AND' (note capital letters are required).
    This may cause problems if you have many files and folders by that
    name, but that should be relatively rare. :-)
""")
MAN_KITE_EXAMPLES = ("""\
    The options <b>--list</b>, <b>--add</b>, <b>--disable</b> and \
<b>--remove</b> can be used to
    manipulate the kites and service definitions in your configuration file,
    if you prefer not to edit it by hand.  Examples:

    <pre>Adding new kites
    $ pagekite.py --add /a/path/ NAME.pagekite.me +indexes
    $ pagekite.py --add 80 OTHER-NAME.pagekite.me

    To display the current configuration
    $ pagekite.py --list

    Disable or delete kites (--add re-enables)
    $ pagekite.py --disable OTHER-NAME.pagekite.me
    $ pagekite.py --remove NAME.pagekite.me</pre>
""")
MAN_FLAGS = ("""\
    Flags are used to tune the behavior of a particular kite, for example
    by enabling access controls or specific features of the built-in HTTP
    server.
""")
MAN_FLAGS_COMMON = ("""\
    +ip</b>/<a>1.2.3.4</a>     __Enable connections only from this IP address.
    +ip</b>/<a>1.2.3</a>       __Enable connections only from this /24 netblock.
""")
MAN_FLAGS_HTTP = ("""\
    +password</b>/<a>name</a>=<a>pass</a>
            Require a username and password (HTTP Basic Authentication)

    +rewritehost</b>    __Rewrite the incoming Host: header.
    +rewritehost</b>=<a>N</a>  __Replace Host: header value with N.
    +rawheaders</b>     __Do not rewrite (or add) any HTTP headers at all.
    +proxyproto</b>     __Use HAProxy's PROXY Protocol (v1) to relay IPs etc.
    +insecure</b>       __Allow access to phpMyAdmin, /admin, etc. (per kite).
""")
MAN_FLAGS_BUILTIN = ("""\
    +indexes        __Enable directory indexes.
    +indexes</b>=<a>all</a>    __Enable directory indexes including hidden (dot-) files.
    +hide           __Obfuscate URLs of shared files.
    +uploads        __Accept file uploads.
    +uploads</b>=<a>RE</a>     __Accept uploads to paths matching regexp RE.
    +ul_filenames</b>=<a>P</a> __Upload naming policy. P = overwrite, keep or rename

    +cgi</b>=<a>list</a>
            A list of extensions, for which files should be treated as
            CGI scripts (example: <tt>+cgi=cgi,pl,sh</tt>).

    +photobackup</b>=<a>password</a>
            Enable built-in PhotoBackup server with the given password.
            See https://photobackup.github.io/ for details.
""")
MAN_OPTIONS = ("""\
    The full power of <b>pagekite.py</b> lies in the numerous options which
    can be specified on the command line or in a configuration file (see below).

    Note that many options, especially the service and domain definitions,
    are additive and if given multiple options the program will attempt to
    obey them all.  Options are processed in order and if they are not
    additive then the last option will override all preceding ones.

    Although <b>pagekite.py</b> accepts a great many options, most of the
    time the program defaults will Just Work.
""")
MAN_OPT_COMMON = ("""\
    --clean         __Skip loading the default configuration file.
    --signup        __Interactively sign up for pagekite.net service.
    --defaults      __Set defaults for use with pagekite.net service.
    --whitelabel=D  __Set defaults for pagekite.net white-labels.
    --whitelabels=D __Set defaults for pagekite.net white-labels (with TLS).
    --nocrashreport __Don't send anonymous crash reports to pagekite.net.
""")
MAN_OPT_BACKEND = ("""\
    --shell         __Run PageKite in an interactive shell.
    --nullui        __Silent UI for scripting. Assumes Yes on all questions.

    --list          __List all configured kites.
    --add           __Add (or enable) the following kites, save config.
    --remove        __Remove the following kites, save config.
    --disable       __Disable the following kites, save config.
    --only          __Disable all but the following kites, save config.

    --insecure      __Allow access to phpMyAdmin, /admin, etc. (global).

    --local</b>=<a>ports</a>   __Configure for local serving only (no remote front-end).
    --watch</b>=<a>N</a>       __Display proxied data (higher N = more verbosity).

    --noproxy       __Ignore system (or config file) proxy settings.

    --proxy</b>=<a>type</a>:<a>server</a>:<a>port</a>,\
 <b>--socksify</b>=<a>server</a>:<a>port</a>,\
 <b>--torify</b>=<a>server</a>:<a>port</a> __
            Connect to the front-ends using SSL, an HTTP proxy, a SOCKS proxy,
            or the Tor anonymity network.  The type can be any of 'ssl', 'http'
            or 'socks5'.  The server name can either be a plain hostname,
            user@hostname or user:password@hostname.  For SSL connections the
            user part may be a path to a client cert PEM file.  If multiple
            proxies are defined, they will be chained one after another.

    --service_on</b>=<a>proto</a>:<a>kitename</a>:<a>host</a>:<a>port</a>:<a>secret</a> __
            Explicit configuration for a service kite.  Generally kites are
            created on the command-line using the service short-hand
            described above, but this syntax is used in the config file.
            The kitename `unknown`, if allowed by the front-end, represents
            a backend of last resort for requests with no other match.

    --authdomain</b>=<a>DNS-suffix</a>,\
 <b>--authdomain</b>=<a>/path/to/app</a>,\
 <b>--authdomain</b>=<a>kite-domain</a>:<a>DNS-suffix</a>,\
 <b>--authdomain</b>=<a>kite-domain</a>:<a>/path/to/app</a> __
            Use <a>DNS-suffix</a> for remote DNS-based authentication of
            incoming tunnel requests, or invoke an external application
            for this purpose.  If no <i>kite-domain</i> is given, use
            this as the default authentication method.  See the section
            below on tunnel authentication for further details.  In order
            for the app path to be recognized as such, it must contain at
            least one / character.

    --auththreads</b>=<a>N</a> __
            Start N threads to process auth requests. Default is 1.

    --authfail_closed</b> __
            If authentication fails, reject tunnel requests. The default is
            to fail open and allow tunnels if the auth checks are broken.


    --service_off</b>=<a>proto</a>:<a>kitename</a>:<a>host</a>:<a>port</a>:<a>secret</a> __
            Same as --service_on, except disabled by default.

    --service_cfg</b>=<a>...</a>, <b>--webpath</b>=<a>...</a> __
            These options are used in the configuration file to store service
            and flag settings (see above). These are both likely to change in
            the near future, so please just pretend you didn't notice them.

    --frontend</b>=<a>host</a>:<a>port</a> __
            Connect to the named front-end server. If this option is repeated,
            multiple connections will be made.

    --frontends</b>=<a>num</a>:<a>dns-name</a>:<a>port</a> __
            Choose <a>num</a> front-ends from the A records of a DNS domain
            name, using the given port number. Default behavior is to probe
            all addresses and use the fastest one.

    --frontends</b>=<a>num</a>:<a>@/path/to/file</a>:<a>port</a> __
            Same as above, except the IP address list will be loaded from
            a file (and reloaded periodically), instead of using DNS.

    --nofrontend</b>=<a>ip</a>:<a>port</a> __
            Never connect to the named front-end server. This can be used to
            exclude some front-ends from auto-configuration.

    --fe_certname</b>=<a>domain</a> __
            Connect using SSL, accepting valid certs for this domain. If
            this option is repeated, any of the named certificates will be
            accepted, but the first will be preferred.

    --fe_nocertcheck</b> __
            Connect using SSL/TLS, but do not verify the remote certificate.
            This is largely insecure but still thwarts passive attacks and
            prevents routers and firewalls from corrupting the PageKite tunnel.

    --ca_certs</b>=<a>/path/to/file</a> __
            Path to your trusted root SSL certificates file.

    --dyndns</b>=<a>X</a> __
            Register changes with DynDNS provider X.  X can either be simply
            the name of one of the 'built-in' providers, or a URL format
            string for ad-hoc updating.

    --keepalive</b>=<a>N</a> __
            Force traffic over idle tunnels every N seconds, to cope with
            firewalls that kill idle TCP connections. Backend only: if set
            to "auto" (the default), the interval will be adjusted
            automatically in response to disconnects.

    --all           __Terminate early if any tunnels fail to register.
    --new           __Don't attempt to connect to any kites' old front-ends.
    --noprobes      __Reject all probes for service state.
""")
MAN_OPT_FRONTEND = ("""\
    --isfrontend    __Enable front-end operation.

    --domain</b>=<a>proto,proto2,pN</a>:<a>domain</a>:<a>secret</a> __
            Accept tunneling requests for the named protocols and specified
            domain, using the given secret.  A * may be used as a wildcard
            for subdomains or protocols. This is for static configurations,
            for dynamic access controls use the `--authdomain` mechanism.
            The domain `unknown`, if configured, represents a backend of
            last resort for incoming requests with no other match.

    --authdomain</b>=<a>DNS-suffix</a>,\
 <b>--authdomain</b>=<a>/path/to/app</a>,\
 <b>--authdomain</b>=<a>kite-domain</a>:<a>DNS-suffix</a>,\
 <b>--authdomain</b>=<a>kite-domain</a>:<a>/path/to/app</a> __
            Use <a>DNS-suffix</a> for remote DNS-based authentication of
            incoming tunnel requests, or invoke an external application
            for this purpose.  If no <i>kite-domain</i> is given, use
            this as the default authentication method.  See the section
            below on tunnel authentication for further details.  In order
            for the app path to be recognized as such, it must contain at
            least one / character.

    --auththreads</b>=<a>N</a> __
            Start N threads to process auth requests. Default is 1.

    --authfail_closed</b> __
            If authentication fails, reject tunnel requests. The default is
            to fail open and allow tunnels if the auth checks are broken.

    --motd</b>=<a>/path/to/motd</a> __
            Send the contents of this file to new back-ends as a
            "message of the day".

    --host</b>=<a>hostname</a> __Listen on the given hostname only.
    --ports</b>=<a>list</a>    __Listen on a comma-separated list of ports.
    --portalias</b>=<a>A:B</a> __Report port A as port B to backends (because firewalls).
    --protos</b>=<a>list</a>   __Accept the listed protocols for tunneling.

    --rawports</b>=<a>list</a> __
            Listen for raw connections these ports. The string '%s'
            allows arbitrary ports in HTTP CONNECT.

    --overload</b>=<a>baseline</a>,\
 <b>--overload_cpu</b>=<a>fraction, 0-1</a>,\
 <b>--overload_mem</b>=<a>fraction, 0-1</a> __
            Enable "overload" calculations, which cause the front-end to
            recommend back-ends go elsewhere if possible, once connection
            counts go above a certain number. The baseline is the initial
            overload level, but it will be adjusted dynamically based on
            load average (CPU use) and memory usage. This will really only
            work well on Linux and if PageKite is the only thing happening
            on the machine. Setting both fractions to 0 disables dynamic
            scaling.

    --overload_file</b>=<a>/path/to/baseline/file</a> __
            Path to a file, the contents of which overrides all overload
            calculations. This can be used to manage load calculations
            using an external process (or by hand, e.g. to prepare for
            maintenance). Note that <i>overload</i> must specify a non-zero
            baseline, otherwise this setting is ignored.

    --ratelimit_ips</b>=<a>IPs/seconds</a>,\
 <b>--ratelimit_ips</b>=<a>kitename</a>:<a>IPs/seconds</a> __
            Limit how many different clients (IPs) can request data from
            a tunnel within a given window of time, e.g. 5/3600. This is
            useful as either a crude form of DDoS mitigation, or as a
            mechanism to make public kite services unusable for phishing.
            Note that limits are enforced per-tunnel (not per kite), and
            tunnels serving multiple kites will use the settings of the
            strictest kite. Limits apply to subdomains as well. A single
            IP may be counted more than once if request headers (such as
            User-Agent) differ.

    --accept_acl_file</b>=<a>/path/to/file</a> __
            Consult an external access control file before accepting an
            incoming connection. Quick'n'dirty for mitigating abuse. The
            format is one rule per line: `rule policy comment` where a
            rule is an IP or regexp and policy is 'allow' or 'deny'.

    --client_acl</b>=<a>policy</a>:<a>regexp</a>,\
 <b>--tunnel_acl</b>=<a>policy</a>:<a>regexp</a> __
            Add a client connection or tunnel access control rule.
            Policies should be 'allow' or 'deny', the regular expression
            should be written to match IPv4 or IPv6 addresses.  If defined,
            access rules are checkd in order and if none matches, incoming
            connections will be rejected.

    --tls_default</b>=<a>name</a> __
            Default name to use for SSL, if SNI (Server Name Indication)
            is missing from incoming HTTPS connections.

    --tls_endpoint</b>=<a>name</a>:<a>/path/to/file</a> __
            Terminate SSL/TLS for a name using key/cert from a file.
""")
MAN_OPT_SYSTEM = ("""\
    --optfile</b>=<a>/path/to/file</a> __
            Read settings from file X. Default is <tt>~/.pagekite.rc</tt>.

    --optdir</b>=<a>/path/to/directory</a> __
            Read settings from <tt>/path/to/directory/*.rc</tt>, in
            lexicographical order.

    --savefile</b>=<a>/path/to/file</a> __
            Saved settings will be written to this file.

    --save          __Save the current configuration to the savefile.

    --settings</b> __
            Dump the current settings to STDOUT, formatted as a configuration
            file would be.

    --nopyopenssl  __Avoid use of the pyOpenSSL library (not in config file)
    --nossl        __Avoid use SSL entirely (not allowed in config file)

    --nozchunks    __Disable zlib tunnel compression.
    --sslzlib      __Enable zlib compression in OpenSSL.
    --buffers</b>=<a>N</a>    __Buffer at most N kB of data before blocking.
    --logfile</b>=<a>F</a>    __Log to file F, <tt>stdio</tt> means standard output.
    --daemonize    __Run as a daemon.
    --runas</b>=<a>U</a>:<a>G</a>    __Set UID:GID after opening our listening sockets.
    --pidfile</b>=<a>P</a>    __Write PID to the named file.
    --errorurl</b>=<a>U</a>   __URL to redirect to when back-ends are not found.
    --errorurl</b>=<a>D:U</a> __Custom error URL for domain D.

    --selfsign</b> __
            Configure the built-in HTTP daemon for HTTPS, first generating a
            new self-signed certificate using <b>openssl</b> if necessary.

    --httpd</b>=<a>X</a>:<a>P</a>,\
 <b>--httppass</b>=<a>X</a>,\
 <b>--pemfile</b>=<a>X</a> __
            Configure the built-in HTTP daemon.  These options are likely to
            change in the near future, please pretend you didn't see them.
""")
MAN_CONFIG_FILES = ("""\
    If you are using <b>pagekite.py</b> as a command-line utility, it will
    load its configuration from a file in your home directory.  The file is
    named <tt>.pagekite.rc</tt> on Unix systems (including Mac OS X), or
    <tt>pagekite.cfg</tt> on Windows.

    If you are using <b>pagekite.py</b> as a system-daemon which starts up
    when your computer boots, it is generally configured to load settings
    from <tt>/etc/pagekite.d/*.rc</tt> (in lexicographical order).

    In both cases, the configuration files contain one or more of the same
    options as are used on the command line, with the difference that at most
    one option may be present on each line, and the parser is more tolerant of
    white-space.  The leading '--' may also be omitted for readability and
    blank lines and lines beginning with '#' are treated as comments.

    <b>NOTE:</b> When using <b>-o</b>, <b>--optfile</b> or <b>--optdir</b> on the command line,
    it is advisable to use <b>--clean</b> to suppress the default configuration.
""")
MAN_SECURITY = ("""\
    Please keep in mind, that whenever exposing a server to the public
    Internet, it is important to think about security. Hacked webservers are
    frequently abused as part of virus, spam or phishing campaigns and in
    some cases security breaches can compromise the entire operating system.

    Some advice:<pre>
       * Switch PageKite off when not using it.
       * Use the built-in access controls and SSL encryption.
       * Leave the firewall enabled unless you have good reason not to.
       * Make sure you use good passwords everywhere.
       * Static content is very hard to hack!
       * Always, always make frequent backups of any important work.</pre>

    Note that as of version 0.5, <b>pagekite.py</b> includes a very basic
    request firewall, which attempts to prevent access to phpMyAdmin and
    other sensitive systems.  If it gets in your way, the <b>+insecure</b>
    flag or <b>--insecure</b> option can be used to turn it off.

    For more, please visit: <https://pagekite.net/support/security/>
""")
MAN_TUNNEL_AUTH = ("""\
    When running <b>pagekite.py</b> as a front-end relay, you can enable
    dynamic authentication of incoming tunnel requests in two ways.

    One uses a DNS-based protocol for delegating authentication to a remote
    server. The nice thing about this, is relays can be deployed without
    any direct access to your user account databases - in particular, a
    zero-knowlege challenge/response protocol is used which means the relay
    never sees the shared secret used to authenticate the kite.

    The second method delegates authentication to an external app; this
    external app can be written in any language you like, as long as it
    implements the following command-line arguments:
    <pre>
      --capabilities     Print a list of capabilities to STDOUT and exit
      --server           Run as a "server", reading queries on STDIN and
                         sending one-line replies to STDOUT.
      --auth <domain>    Return JSON formatted auth and quota details
      --zk-auth <query>  Implement the DNS-based zero-knowlege protocol
    </pre>
    The recognized capabilities are SERVER, ZK-AUTH and AUTH. One of AUTH
    or ZK-AUTH is required.

    The JSON `--auth` responses should be dictionaries which have at least
    one element, `secret` or `error`. The secret is the shared secret to
    be used to authenticate the tunnel. The dictionary may also contain
    advisory quota values (`quota_kb`, `quota_days` and `quota_conns`), and
    IP rate limiting parameters (`ips_per_sec-ips` and `ips_per_sec-secs`).

    The source distribution of <b>pagekite.py</b> includes a script named
    `demo_auth_app.py` which implements this protocol.
""")
MAN_LICENSE = ("""\
    Copyright 2010-2020, the Beanstalks Project ehf. and Bjarni R. Einarsson.

    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or (at
    your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public
    License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see: <http://www.gnu.org/licenses/>
""")
MAN_BUGS = ("""\
    Using <b>pagekite.py</b> as a front-end relay with the native Python SSL
    module may result in poor performance.  Please use the pyOpenSSL wrappers
    instead.
""")
MAN_SEE_ALSO = ("""\
    lapcat(1), <http://pagekite.org/>, <https://pagekite.net/>
""")
MAN_CREDITS = ("""\
    <pre>- Bjarni R. Einarsson <http://bre.klaki.net/>
    - The Beanstalks Project ehf. <https://pagekite.net/company/>
    - The Rannis Technology Development Fund <http://www.rannis.is/>
    - Joar Wandborg <http://wandborg.se/>
    - Luc-Pierre Terral</pre>
""")

MANUAL_TOC = (
  ('SH', 'Name', MAN_NAME),
  ('SH', 'Synopsis', MAN_SYNOPSIS),
  ('SH', 'Description', MAN_DESCRIPTION),
  ('SH', 'Basic usage', MAN_EXAMPLES),
  ('SH', 'Services and kites', MAN_KITES),
  ('SH', 'Kite configuration', MAN_KITE_EXAMPLES),
  ('SH', 'Flags', MAN_FLAGS),
  ('SS', 'Common flags', MAN_FLAGS_COMMON),
  ('SS', 'HTTP protocol flags', MAN_FLAGS_HTTP),
  ('SS', 'Built-in HTTPD flags', MAN_FLAGS_BUILTIN),
  ('SH', 'Options', MAN_OPTIONS),
  ('SS', 'Common options', MAN_OPT_COMMON),
  ('SS', 'Back-end options', MAN_OPT_BACKEND),
  ('SS', 'Front-end options', MAN_OPT_FRONTEND),
  ('SS', 'System options', MAN_OPT_SYSTEM),
  ('SH', 'Configuration files', MAN_CONFIG_FILES),
  ('SH', 'Security', MAN_SECURITY),
  ('SH', 'Tunnel Request Authentication', MAN_TUNNEL_AUTH),
  ('SH', 'Bugs', MAN_BUGS),
  ('SH', 'See Also', MAN_SEE_ALSO),
  ('SH', 'Credits', MAN_CREDITS),
  ('SH', 'Copyright and license', MAN_LICENSE),
)

HELP_SHELL = ("""\
    Press ENTER to fly your kites, CTRL+C to quit or give some arguments to
    accomplish a more specific task.
""")
HELP_KITES = ("""\
""")
HELP_TOC = (
  ('about',    'About PageKite',                        MAN_DESCRIPTION),
  ('basics',   'Basic usage examples',                  MAN_EXAMPLES),
  ('kites',    'Services and kites',                    MAN_KITES),
  ('config',   'Adding, disabling or removing kites',   MAN_KITE_EXAMPLES),
  ('flags',    'Service flags',              '\n'.join([MAN_FLAGS,
                                                        MAN_FLAGS_COMMON,
                                                        MAN_FLAGS_HTTP,
                                                        MAN_FLAGS_BUILTIN])),
  ('files',    'Where are the config files?',           MAN_CONFIG_FILES),
  ('security', 'A few words about security.',           MAN_SECURITY),
  ('credits',  'License and credits',        '\n'.join([MAN_LICENSE,
                                                        'CREDITS:',
                                                        MAN_CREDITS])),
  ('manual', 'The complete manual.  See also: http://pagekite.net/man/', None)
)


def HELP(args):
  name = title = text = ''
  if args:
    what = args[0].strip().lower()
    for name, title, text in HELP_TOC:
      if name == what:
        break
  if name == 'manual':
    text = DOC()
  elif not text:
    text = ''.join([
      'Type `help TOPIC` to to read about one of these topics:\n\n',
      ''.join(['  %-10.10s %s\n' % (n, t) for (n, t, x) in HELP_TOC]),
      '\n',
      HELP_SHELL
    ])
  return unindent(clean_text(text))


def clean_text(text):
  return re.sub('</?(tt|i)>', '`',
                re.sub('</?(a|b|pre)>', '', text.replace(' __', '   ')))

def unindent(text):
  return re.sub('(?m)^    ', '', text)


def MINIDOC():
  return ("""\
>>> Welcome to pagekite.py v%s!

%s
    To sign up with PageKite.net or get advanced instructions:
    $ pagekite.py --signup
    $ pagekite.py --help

    If you request a kite which does not exist in your configuration file,
    the program will offer to help you sign up with https://pagekite.net/
    and create it.  Pick a name, any name!\
""") % (APPVER, clean_text(MAN_EXAMPLES))


def DOC():
  doc = ''
  for h, section, text in MANUAL_TOC:
    doc += '%s\n\n%s\n' % (h == 'SH' and section.upper() or '  '+section,
                           clean_text(text))
  return doc


def MAN(pname=None):
  lastchange = float(os.environ.get('SOURCE_DATE_EPOCH',
                                    os.path.getmtime(sys.argv[0])))
  man = ("""\
.\\" This man page is autogenerated from the pagekite.py built-in manual.
.TH PAGEKITE "1" "%s" "https://pagekite.net/" "Awesome Commands"
.nh
.ad l
""") % ts_to_iso(lastchange).split('T')[0]
  for h, section, text in MANUAL_TOC:
    man += ('.%s %s\n\n%s\n\n'
            ) % (h, h == 'SH' and section.upper() or section,
                 re.sub('\n +', '\n', '\n'+text.strip())
                   .replace('\n--', '\n.TP\n\\fB--')
                   .replace('\n+', '\n.TP\n\\fB+')
                   .replace(' __', '\\fR\n')
                   .replace('-', '\\-')
                   .replace('<pre>', '\n.nf\n').replace('</pre>', '\n.fi\n')
                   .replace('<b>', '\\fB').replace('</b>', '\\fR')
                   .replace('<a>', '\\fI').replace('</a>', '\\fR')
                   .replace('<i>', '\\fI').replace('</i>', '\\fR')
                   .replace('<tt>', '\\fI').replace('</tt>', '\\fR')
                   .replace('\\fR\\fR\n', '\\fR'))
  if pname:
    man = man.replace('pagekite.py', pname)
  return man


def MARKDOWN(pname=None):
  mkd = ''
  for h, section, text in MANUAL_TOC:
     if h == 'SH':
       h = '##'
     else:
       h = '###'
     mkd += ('%s %s %s\n%s\n\n'
            ) % (h, section, h,
                 re.sub('(</[aib]>|`)</b>', '\\1',
                  re.sub(' +<br />([A-Z0-9])', '</b>  \n     \\1',
                   re.sub('\n        ', '\n     ',
                    re.sub('\n    ', '\n', '\n'+text.strip()))
                     .replace(' __', ' <br />')
                     .replace('\n--', '\n   * <b>--')
                     .replace('\n+', '\n   * <b>+')
                     .replace('<a>', '`').replace('</a>', '`')
                     .replace('<tt>', '`').replace('</tt>', '`'))))
  if pname:
    mkd = mkd.replace('pagekite.py', pname)
  return mkd


if __name__ == '__main__':
  import sys
  if '--nopy' in sys.argv:
    pname = 'pagekite'
  else:
    pname = None

  if '--man' in sys.argv:
    print(MAN(pname))
  elif '--markdown' in sys.argv:
    print(MARKDOWN(pname))
  elif '--minidoc' in sys.argv:
    print(MINIDOC())
  else:
    print(DOC())



#!/usr/bin/python -u

from __future__ import absolute_import
from __future__ import print_function

#
# droiddemo.py, Copyright 2010-2013, The Beanstalks Project ehf.
#                                    http://beanstalks-project.net/
#
# This is a proof-of-concept PageKite enabled HTTP server for Android.
# It has been developed and tested in the SL4A Python environment.
#
DOMAIN='phone.bre.pagekite.me'
SECRET='ba4e5430'
SOURCE='/sdcard/sl4a/scripts/droiddemo.py'
#
#############################################################################
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#############################################################################
#

import six
from six.moves.urllib import unquote
from six.moves.urllib.parse import urlparse

import android
import pagekite
import os


class UiRequestHandler(pagekite.UiRequestHandler):

  CAMERA_PATH = '/sdcard/dcim/.thumbnails'
  HOME = ('<html><head>\n'
          '<script type=text/javascript>'
           'lastImage = "";'
           'function getImage() {'
            'xhr = new XMLHttpRequest();'
            'xhr.open("GET", "/latest-image.txt", true);'
            'xhr.onreadystatechange = function() {'
             'if (xhr.readyState == 4) {'
              'if (xhr.responseText && xhr.responseText != lastImage) {'
               'document.getElementById("i").src = lastImage = xhr.responseText;'
              '}'
              'setTimeout("getImage()", 2000);'
             '}'
            '};'
           'xhr.send(null);'
           '}'
          '</script>\n'
          '</head><body onLoad="getImage();" style="text-align: center;">\n'
          '<h1>Android photos!</h1>\n'
          '<img id=i height=80% src="http://www.android.com/images/opensourceproject.gif">\n'
          '<br><a href="/droiddemo.py">source code</a>'
          '| <a href="/status.html">kite status</a>\n'
          '</body></head>')

  def listFiles(self):
    mtimes = {}
    for item in os.listdir(self.CAMERA_PATH):
      iname = '%s/%s' % (self.CAMERA_PATH, item)
      if iname.endswith('.jpg'):
        mtimes[iname] = os.path.getmtime(iname)

    files = list(six.iterkeys(mtimes))
    files.sort(key=lambda iname: mtimes[iname])
    return files

  def do_GET(self):
    (scheme, netloc, path, params, query, frag) = urlparse(self.path)

    p = unquote(path)
    if p.endswith('.jpg') and p.startswith(self.CAMERA_PATH) and ('..' not in p):
      try:
        jpgfile = open(p)
        self.send_response(200)
        self.send_header('Content-Type', 'image/jpeg')
        self.send_header('Content-Length', '%s' % os.path.getsize(p))
        self.send_header('Cache-Control', 'max-age: 36000')
        self.send_header('Expires', 'Sat, 1 Jan 2011 12:00:00 GMT')
        self.send_header('Last-Modified', 'Wed, 1 Sep 2011 12:00:00 GMT')
        self.end_headers()
        data = jpgfile.read() 
        while data:
          try:
            sent = self.wfile.write(data[0:15000])
            data = data[15000:]
          except Exception:
            pass
        return
 
      except Exception as e:
        print('%s' % e)
        pass 

    if path == '/latest-image.txt':
      flist = self.listFiles()
      self.begin_headers(200, 'text/plain')
      self.end_headers()
      self.wfile.write(flist[-1])
      return
    elif path == '/droiddemo.py':
      try:
        pyfile = open(SOURCE)
        self.begin_headers(200, 'text/plain')
        self.end_headers()
        self.wfile.write(pyfile.read().replace(SECRET, 'mysecret'))
      except IOError as e:
        self.begin_headers(404, 'text/plain')
        self.end_headers()
        self.wfile.write('Could not read %s: %s' % (SOURCE, e))
      return
    elif path == '/':
      self.begin_headers(200, 'text/html')
      self.end_headers()
      self.wfile.write(self.HOME)
      return

    return pagekite.UiRequestHandler.do_GET(self)


class DroidKite(pagekite.PageKite):
  def __init__(self, droid):
    pagekite.PageKite.__init__(self)
    self.droid = droid
    self.ui_request_handler = UiRequestHandler


def Start(host, secret):
  ds = DroidKite(android.Android())
  ds.Configure(['--defaults',
                '--httpd=localhost:9999',
                '--backend=http:%s:localhost:9999:%s' % (host, secret)])
  ds.Start()


Start(DOMAIN, SECRET)



"""
A basic tool for processing and parsing the Pagekite logs. This class
doesn't actually do anything much, it's meant for subclassing.
"""

from __future__ import absolute_import
from __future__ import print_function

##############################################################################
LICENSE = """\
This file is part of pagekite.py.
Copyright 2010-2020, the Beanstalks Project ehf. and Bjarni Runar Einarsson

This program is free software: you can redistribute it and/or modify it under
the terms of the  GNU  Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful,  but  WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see: <http://www.gnu.org/licenses/>
"""
##############################################################################

import os
import sys
import time


class PageKiteLogParser(object):
  def __init__(self):
    pass

  def ParseLine(self, line, data=None):
    try:
      if data is None: data = {}
      for word in line.split('; '):
        key, val = word.split('=', 1);
        data[key] = val
      return data
    except Exception:
      return {'raw': '%s' % line}

  def ProcessData(self, data):
    print('%s' % data)

  def ProcessLine(self, line, data=None):
    self.ProcessData(self.ParseLine(line, data))

  def Follow(self, fd, filename):
    # Record last position...
    pos = fd.tell()

    try:
      if os.stat(filename).st_size < pos:
        # Re-open log-file if it's been rotated/trucated
        new_fd = open(filename, 'r')
        fd.close()
        return new_fd
    except (OSError, IOError) as e:
      # Failed to stat or open new file, just try again later.
      pass

    # Sleep a bit and then try to read some more
    time.sleep(1)
    fd.seek(pos)
    return fd

  def ReadLog(self, filename=None, after=None, follow=False):
    if filename is not None:
      fd = open(filename, 'r')
    else:
      fd = sys.stdin

    first = True
    while first or follow:
      for line in fd:
        if line.endswith('\n'):
          data = self.ParseLine(line.strip())
          if after is None or ('ts' in data and int(data['ts'], 16) > after):
            self.ProcessData(data)
        else:
          fd.seek(fd.tell() - len(line))
          break

      if follow: fd = self.Follow(fd, filename)
      first = False

  def ReadSyslog(self, filename, pname='pagekite.py', after=None, follow=False):
    fd = open(filename, 'r')
    tag = ' %s[' % pname
    first = True
    while first or follow:
      for line in fd:
        if line.endswith('\n'):
          try:
            parts = line.split(':', 3)
            if parts[2].find(tag) > -1:
              data = self.ParseLine(parts[3].strip())
              if after is None or int(data['ts'], 16) > after:
                self.ProcessData(data)
          except ValueError as e:
            pass
        else:
          fd.seek(fd.tell() - len(line))
          break

      if follow: fd = self.Follow(fd, filename)
      first = False

class PageKiteLogTracker(PageKiteLogParser):
  def __init__(self):
    PageKiteLogParser.__init__(self)
    self.streams = {}

  def ProcessRestart(self, data):
    # Program just restarted, discard streams state.
    self.streams = {}

  def ProcessBandwidthRead(self, stream, data):
    stream['read'] += int(data['read'])

  def ProcessBandwidthWrote(self, stream, data):
    stream['wrote'] += int(data['wrote'])

  def ProcessError(self, stream, data):
    stream['err'] = data['err']

  def ProcessEof(self, stream, data):
    del self.streams[stream['id']]

  def ProcessNewStream(self, stream, data):
    self.streams[stream['id']] = stream
    stream['read'] = 0
    stream['wrote'] = 0

  def ProcessData(self, data):
    if 'id' in data:
      # This is info about a specific stream...
      sid = data['id']

      if 'proto' in data and 'domain' in data and sid not in self.streams:
        self.ProcessNewStream(data, data)

      if sid in self.streams:
        stream = self.streams[sid]

        if 'err' in data: self.ProcessError(stream, data)
        if 'read' in data: self.ProcessBandwidthRead(stream, data)
        if 'wrote' in data: self.ProcessBandwidthWrote(stream, data)
        if 'eof' in data: self.ProcessEof(stream, data)

    elif 'started' in data and 'version' in data:
      self.ProcessRestart(data)


class DebugPKLT(PageKiteLogTracker):

  def ProcessRestart(self, data):
    PageKiteLogTracker.ProcessRestart(self, data)
    print('RESTARTED %s' % data)

  def ProcessNewStream(self, stream, data):
    PageKiteLogTracker.ProcessNewStream(self, stream, data)
    print('[%s] NEW %s' % (stream['id'], data))

  def ProcessBandwidthRead(self, stream, data):
    PageKiteLogTracker.ProcessBandwidthRead(self, stream, data)
    print('[%s] BWR  %s' % (stream['id'], data))

  def ProcessBandwidthWrote(self, stream, data):
    PageKiteLogTracker.ProcessBandwidthWrote(self, stream, data)
    print('[%s] BWW %s' % (stream['id'], data))

  def ProcessError(self, stream, data):
    PageKiteLogTracker.ProcessError(self, stream, data)
    print('[%s] ERR %s' % (stream['id'], data))

  def ProcessEof(self, stream, data):
    PageKiteLogTracker.ProcessEof(self, stream, data)
    print('[%s] EOF %s' % (stream['id'], data))


if __name__ == '__main__':
  sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
  if len(sys.argv) > 2:
    DebugPKLT().ReadSyslog(sys.argv[1], pname=sys.argv[2])
  else:
    DebugPKLT().ReadLog(sys.argv[1])




"""
Constants and global program state.
"""
##############################################################################

from __future__ import absolute_import

LICENSE = """\
This file is part of pagekite.py.
Copyright 2010-2020, the Beanstalks Project ehf. and Bjarni Runar Einarsson

This program is free software: you can redistribute it and/or modify it under
the terms of the  GNU  Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful,  but  WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see: <http://www.gnu.org/licenses/>
"""
##############################################################################
import random
import time

PROTOVER = '0.8'
APPVER = '1.5.2.201011'
AUTHOR = 'Bjarni Runar Einarsson, http://bre.klaki.net/'
WWWHOME = 'https://pagekite.net/'
LICENSE_URL = 'http://www.gnu.org/licenses/agpl.html'

MAGIC_PREFIX = '/~:PageKite:~/'
MAGIC_PATH = '%sv%s' % (MAGIC_PREFIX, PROTOVER)
MAGIC_PATHS = (MAGIC_PATH, '/Beanstalk~Magic~Beans/0.2')
MAGIC_UUID = '%x-%x-%s' % (random.randint(0, 0xfffffff), int(time.time()), APPVER)

SERVICE_PROVIDER = 'PageKite.net'
SERVICE_DOMAINS = ('pagekite.me', '302.is', 'testing.is', 'kazz.am')
SERVICE_DOMAINS_SIGNUP = ('pagekite.me',)
SERVICE_XMLRPC = 'http://pagekite.net/xmlrpc/'
SERVICE_TOS_URL = 'https://pagekite.net/humans.txt'
SERVICE_CERTS = ['b5p.us', 'frontends.b5p.us', 'pagekite.net', 'pagekite.me',
                 'pagekite.com', 'pagekite.org', 'testing.is', '302.is']

# Places to search for the CA Certificate bundle
OS_CA_CERTS = (
  "/etc/pki/tls/certs/ca-bundle.crt",        # Fedora/RHEL
  "/etc/ssl/certs/ca-certificates.crt",      # Debian/Ubuntu/Gentoo etc.
  "/etc/ssl/ca-bundle.pem",                  # OpenSUSE
  "/etc/pki/tls/cacert.pem",                 # OpenELEC
  "/etc/ssl/cert.pem",                       # OpenBSD
  "/usr/local/share/certs/ca-root-nss.crt",  # FreeBSD/DragonFly
  "/usr/local/etc/openssl/cert.pem",         # OS X (Homebrew)
  "/opt/local/etc/openssl/cert.pem",         # OS X (Ports?)
  "/system/etc/security/cacerts")            # Android

CURL_CA_CERTS = 'https://curl.haxx.se/ca/cacert.pem'


DEFAULT_CHARSET = 'utf-8'
DEFAULT_BUFFER_MAX = 1024

AUTH_ERRORS           = '255.255.255.'
AUTH_ERR_USER_UNKNOWN = '.0'
AUTH_ERR_INVALID      = '.1'
AUTH_QUOTA_MAX        = '255.255.254.255'

VIRTUAL_PN = 'virtual'
CATCHALL_HN = 'unknown'
LOOPBACK_HN = 'loopback'
LOOPBACK_FE = LOOPBACK_HN + ':1'
LOOPBACK_BE = LOOPBACK_HN + ':2'
LOOPBACK = {'FE': LOOPBACK_FE, 'BE': LOOPBACK_BE}

# This is how many bytes we are willing to read per cycle.
MAX_READ_BYTES = (16 * 1024) - 128  # Under 16kB, because OpenSSL
MAX_READ_TUNNEL_X = 3.1             # 3x above, + fudge factor

# Higher values save CPU and prevent individual tunnels
# from hogging all our resources, but hurt latency and
# reduce per-tunnel throughput.
SELECT_LOOP_MIN_MS = 5

# Re-evaluate our choice of frontends every 45-60 minutes.
FE_PING_INTERVAL = (45 * 60) + random.randint(0, 900)

# This is a global count of disconnect errors; we use this
# to adjust the ping interval over time.
DISCONNECTS = []

PING_INTERVAL_MIN     = 20
PING_INTERVAL         = 116  # Not quite 2 minutes... :-)
PING_INTERVAL_DEFAULT = 116
PING_INTERVAL_MOBILE  = 1800
PING_INTERVAL_MAX     = 1800
PING_GRACE_DEFAULT    = 40
PING_GRACE_MIN        = 5

WEB_POLICY_DEFAULT = 'default'
WEB_POLICY_PUBLIC = 'public'
WEB_POLICY_PRIVATE = 'private'
WEB_POLICY_OTP = 'otp'
WEB_POLICIES = (WEB_POLICY_DEFAULT, WEB_POLICY_PUBLIC,
                WEB_POLICY_PRIVATE, WEB_POLICY_OTP)

WEB_INDEX_ALL = 'all'
WEB_INDEX_ON = 'on'
WEB_INDEX_OFF = 'off'
WEB_INDEXTYPES = (WEB_INDEX_ALL, WEB_INDEX_ON, WEB_INDEX_OFF)

BE_PROTO = 0
BE_PORT = 1
BE_DOMAIN = 2
BE_BHOST = 3
BE_BPORT = 4
BE_SECRET = 5
BE_STATUS = 6

BE_STATUS_REMOTE_SSL   = 0x0010000
BE_STATUS_OK           = 0x0001000
BE_STATUS_ERR_DNS      = 0x0000100
BE_STATUS_ERR_BE       = 0x0000010
BE_STATUS_ERR_TUNNEL   = 0x0000001
BE_STATUS_ERR_ANY      = 0x0000fff
BE_STATUS_UNKNOWN      = 0
BE_STATUS_DISABLED     = 0x8000000
BE_STATUS_DISABLE_ONCE = 0x4000000
BE_INACTIVE = (BE_STATUS_DISABLED, BE_STATUS_DISABLE_ONCE)

BE_NONE = ['', '', None, None, None, '', BE_STATUS_UNKNOWN]

DYNDNS = {
  'pagekite.net': ('http://up.pagekite.net/'
                   '?hostname=%(domain)s&myip=%(ips)s&sign=%(sign)s'),
  'beanstalks.net': ('http://up.b5p.us/'
                     '?hostname=%(domain)s&myip=%(ips)s&sign=%(sign)s'),
  'whitelabel': ('http://dnsup.%s/'
                 '?hostname=%%(domain)s&myip=%%(ips)s&sign=%%(sign)s'),
  'whitelabels': ('https://dnsup.%s/'
                  '?hostname=%%(domain)s&myip=%%(ips)s&sign=%%(sign)s'),
  'dyndns.org': ('https://%(user)s:%(pass)s@members.dyndns.org'
                 '/nic/update?wildcard=NOCHG&backmx=NOCHG'
                 '&hostname=%(domain)s&myip=%(ip)s'),
  'no-ip.com': ('https://%(user)s:%(pass)s@dynupdate.no-ip.com'
                '/nic/update?hostname=%(domain)s&myip=%(ip)s'),
}

# Create our service-domain matching regexp
import re
SERVICE_DOMAIN_RE = re.compile('\.(' + '|'.join(SERVICE_DOMAINS) + ')$')
SERVICE_SUBDOMAIN_RE = re.compile(r'^([A-Za-z0-9_-]+\.)*[A-Za-z0-9_-]+$')


class ConfigError(Exception):
  """This error gets thrown on configuration errors."""

class ConnectError(Exception):
  """This error gets thrown on connection errors."""

class BugFoundError(Exception):
  """Throw this anywhere a bug is detected and we want a crash."""


##[ Ugly fugly globals ]#######################################################

# The global Yamon is used for measuring internal state for monitoring
gYamon = None

# Status of our buffers...
buffered_bytes = [0]



"""
This is a user interface class which communicates over a pipe or socket.
"""

from __future__ import absolute_import
from __future__ import print_function

##############################################################################
LICENSE = """\
This file is part of pagekite.py.
Copyright 2010-2020, the Beanstalks Project ehf. and Bjarni Runar Einarsson

This program is free software: you can redistribute it and/or modify it under
the terms of the  GNU  Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful,  but  WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see: <http://www.gnu.org/licenses/>
"""
##############################################################################

import re
import sys
import time
import threading

from pagekite.compat import *
from pagekite.common import *
from pagekite.proto.conns import Tunnel

from .nullui import NullUi

class RemoteUi(NullUi):
  """Stdio based user interface for interacting with other processes."""

  DAEMON_FRIENDLY = True
  ALLOWS_INPUT = True
  WANTS_STDERR = True
  EMAIL_RE = re.compile(r'^[a-z0-9!#$%&\'\*\+\/=?^_`{|}~-]+'
                         '(?:\.[a-z0-9!#$%&\'*+/=?^_`{|}~-]+)*@'
                         '(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)*'
                         '(?:[a-zA-Z]{2,4}|museum)$')

  def __init__(self, welcome=None, wfile=sys.stderr, rfile=sys.stdin):
    NullUi.__init__(self, welcome=welcome, wfile=wfile, rfile=rfile)
    self.CLEAR = ''
    self.NORM = self.WHITE = self.GREY = self.GREEN = self.YELLOW = ''
    self.BLUE = self.RED = self.MAGENTA = self.CYAN = ''

  def StartListingBackEnds(self):
    self.write('begin_be_list\n')

  def EndListingBackEnds(self):
    self.write('end_be_list\n')

  def NotifyBE(self, bid, be, has_ssl, dpaths,
               is_builtin=False, fingerprint=None, now=None):
    domain = be[BE_DOMAIN]
    port = be[BE_PORT]
    proto = be[BE_PROTO]
    prox = (proto == 'raw') and ' (HTTP proxied)' or ''
    if proto == 'raw' and port in ('22', 22): proto = 'ssh'
    url = '%s://%s%s' % (proto, domain, port and (':%s' % port) or '')

    message = (' be_status:'
               ' status=%x; bid=%s; domain=%s; port=%s; proto=%s;'
               ' bhost=%s; bport=%s%s%s%s'
               '\n') % (int(be[BE_STATUS]), bid, domain, port, proto,
                        be[BE_BHOST], be[BE_BPORT],
                        has_ssl and '; ssl=1' or '',
                        is_builtin and '; builtin=1' or '',
                        fingerprint and ('; fingerprint=%s' % fingerprint) or '')
    self.write(message)

    for path in dpaths:
      message = (' be_path: domain=%s; port=%s; path=%s; policy=%s; src=%s\n'
                 ) % (domain, port or 80, path,
                      dpaths[path][0], dpaths[path][1])
      self.write(message)

  def Notify(self, message, prefix=' ',
             popup=False, color=None, now=None, alignright=''):
    message = '%s' % message
    self.write('notify: %s\n' % message)

  def NotifyMOTD(self, frontend, message):
    self.write('motd: %s %s\n' % (frontend,
                                        message.replace('\n', '  ')))

  def Status(self, tag, message=None, color=None):
    self.status_tag = tag
    self.status_msg = '%s' % (message or self.status_msg)
    if message:
      self.write('status_msg: %s\n' % message)
    if tag:
      self.write('status_tag: %s\n' % tag)

  def Welcome(self, pre=None):
    self.write('welcome: %s\n' % (pre or '').replace('\n', '  '))

  def StartWizard(self, title):
    self.write('start_wizard: %s\n' % title)

  def Retry(self):
    self.tries -= 1
    if self.tries < 0:
      raise Exception('Too many tries')
    return self.tries

  def EndWizard(self, quietly=False):
    self.write('end_wizard: %s\n' % (quietly and 'quietly' or 'done'))

  def Spacer(self):
    pass

  def AskEmail(self, question, default=None, pre=[],
               wizard_hint=False, image=None, back=None, welcome=True):
    while self.Retry():
      self.write('begin_ask_email\n')
      if pre:
        self.write(' preamble: %s\n' % '\n'.join(pre).replace('\n', '  '))
      if default:
        self.write(' default: %s\n' % default)
      self.write(' question: %s\n' % (question or '').replace('\n', '  '))
      self.write(' expect: email\n')
      self.write('end_ask_email\n')

      answer = self.rfile.readline().strip()
      if self.EMAIL_RE.match(answer): return answer
      if back is not None and answer == 'back': return back

  def AskLogin(self, question, default=None, email=None, pre=None,
               wizard_hint=False, image=None, back=None):
    while self.Retry():
      self.write('begin_ask_login\n')
      if pre:
        self.write(' preamble: %s\n' % '\n'.join(pre).replace('\n', '  '))
      if email:
        self.write(' default: %s\n' % email)
      self.write(' question: %s\n' % (question or '').replace('\n', '  '))
      self.write(' expect: email\n')
      self.write(' expect: password\n')
      self.write('end_ask_login\n')

      answer_email = self.rfile.readline().strip()
      if back is not None and answer_email == 'back': return back

      answer_pass = self.rfile.readline().strip()
      if back is not None and answer_pass == 'back': return back

      if self.EMAIL_RE.match(answer_email) and answer_pass:
        return (answer_email, answer_pass)

  def AskYesNo(self, question, default=None, pre=[], yes='Yes', no='No',
               wizard_hint=False, image=None, back=None):
    while self.Retry():
      self.write('begin_ask_yesno\n')
      if yes:
        self.write(' yes: %s\n' % yes)
      if no:
        self.write(' no: %s\n' % no)
      if pre:
        self.write(' preamble: %s\n' % '\n'.join(pre).replace('\n', '  '))
      if default:
        self.write(' default: %s\n' % default)
      self.write(' question: %s\n' % (question or '').replace('\n', '  '))
      self.write(' expect: yesno\n')
      self.write('end_ask_yesno\n')

      answer = self.rfile.readline().strip().lower()
      if back is not None and answer == 'back': return back
      if answer in ('y', 'n'): return (answer == 'y')
      if answer == str(default).lower(): return default

  def AskKiteName(self, domains, question, pre=[], default=None,
                  wizard_hint=False, image=None, back=None):
    while self.Retry():
      self.write('begin_ask_kitename\n')
      if pre:
        self.write(' preamble: %s\n' % '\n'.join(pre).replace('\n', '  '))
      for domain in domains:
        self.write(' domain: %s\n' % domain)
      if default:
        self.write(' default: %s\n' % default)
      self.write(' question: %s\n' % (question or '').replace('\n', '  '))
      self.write(' expect: kitename\n')
      self.write('end_ask_kitename\n')

      answer = self.rfile.readline().strip().lower()
      if back is not None and answer == 'back': return back
      if answer:
        for d in domains:
          if answer.endswith(d) or answer.endswith(d): return answer
        return answer+domains[0]

  def AskBackends(self, kitename, protos, ports, rawports, question, pre=[],
                  default=None, wizard_hint=False, image=None, back=None):
    while self.Retry():
      self.write('begin_ask_backends\n')
      if pre:
        self.write(' preamble: %s\n' % '\n'.join(pre).replace('\n', '  '))
      count = 0
      if self.server_info:
        protos = self.server_info[Tunnel.S_PROTOS]
        ports = self.server_info[Tunnel.S_PORTS]
        rawports = self.server_info[Tunnel.S_RAW_PORTS]
      self.write(' kitename: %s\n' % kitename)
      self.write(' protos: %s\n' % ', '.join(protos))
      self.write(' ports: %s\n' % ', '.join(ports))
      self.write(' rawports: %s\n' % ', '.join(rawports))
      if default:
        self.write(' default: %s\n' % default)
      self.write(' question: %s\n' % (question or '').replace('\n', '  '))
      self.write(' expect: backends\n')
      self.write('end_ask_backends\n')

      answer = self.rfile.readline().strip().lower()
      if back is not None and answer == 'back': return back
      return answer

  def AskMultipleChoice(self, choices, question, pre=[], default=None,
                        wizard_hint=False, image=None, back=None):
    while self.Retry():
      self.write('begin_ask_multiplechoice\n')
      if pre:
        self.write(' preamble: %s\n' % '\n'.join(pre).replace('\n', '  '))
      count = 0
      for choice in choices:
        count += 1
        self.write(' choice_%d: %s\n' % (count, choice))
      if default:
        self.write(' default: %s\n' % default)
      self.write(' question: %s\n' % (question or '').replace('\n', '  '))
      self.write(' expect: choice_index\n')
      self.write('end_ask_multiplechoice\n')

      answer = self.rfile.readline().strip().lower()
      try:
        ch = int(answer)
        if ch > 0 and ch <= len(choices): return ch
      except:
        pass
      if back is not None and answer == 'back': return back

  def Tell(self, lines, error=False, back=None):
    dialog = error and 'error' or 'message'
    self.write('tell_%s: %s\n' % (dialog, '  '.join(lines)))

  def Working(self, message):
    self.write('working: %s\n' % message)


class PageKiteThread(threading.Thread):
  daemon = True

  def __init__(self, startup_args=None, debug=False):
    threading.Thread.__init__(self)
    self.pk = None
    self.pk_readlock = threading.Condition()
    self.gui_readlock = threading.Condition()
    self.debug = debug
    self.reset()

  def reset(self):
    self.pk_incoming = []
    self.pk_eof = False
    self.gui_incoming = ''
    self.gui_eof = False

  # These routines are used by the PageKite UI, to communicate with us...
  def readline(self):
    with self.pk_readlock:
      while (not self.pk_incoming) and (not self.pk_eof):
        self.pk_readlock.wait()
      if self.pk_incoming:
        line = self.pk_incoming.pop(0)
      else:
        line = ''
      if self.debug:
        print('>>PK>> %s' % line.strip())
      return line

  def write(self, data):
    if self.debug:
      print('>>GUI>> %s' % data.strip())
    with self.gui_readlock:
      if data:
        self.gui_incoming += data
      else:
        self.gui_eof = True
      self.gui_readlock.notify()

  # And these are used by the GUI, to communicate with PageKite.
  def recv(self, bytecount):
    with self.gui_readlock:
      while (len(self.gui_incoming) < bytecount) and (not self.gui_eof):
        self.gui_readlock.wait()
      data = self.gui_incoming[0:bytecount]
      self.gui_incoming = self.gui_incoming[bytecount:]
      return data

  def send(self, data):
    if not data.endswith('\n') and data != '':
      raise ValueError('Please always send whole lines')
    if self.debug:
      print('<<PK<< %s' % data.strip())
    with self.pk_readlock:
      if data:
        self.pk_incoming.append(data)
      else:
        self.pk_eof = True
      self.pk_readlock.notify()

  def sendall(self, data):
    return self.send(data)

  def close(self):
    self.send('')
    self.write('')

  def setup_comms(self, pkobj):
    self.pk = pkobj
    pkobj.ui_wfile = pkobj.ui.wfile = self
    pkobj.ui_rfile = pkobj.ui.rfile = self

  def run(self):
    raise Exception('Unimplemented')


class PageKiteRestarter(PageKiteThread):

  def __init__(self, startup_args=None):
    PageKiteThread.__init__(self)
    self.pk_startup_args = startup_args
    self.looping = False
    self.stopped = True

  def config_wrapper(self, pkobj):
    old_argv = sys.argv[:]

    # Remove invalid arguments that break us.
    for evil in ('--nullui', '--basicui', '--friendly'):
      if evil in sys.argv:
        sys.argv.remove(evil)

    if self.pk_startup_args:
      sys.argv[1:1] = self.pk_startup_args[:]
      self.pk_startup_args = None
    try:
      try:
        self.setup_comms(pkobj)
        return self.configure(pkobj)
      except:
        self.pk = None
        raise
    finally:
      sys.argv = old_argv[:]

  def run(self):
    self.looping = True
    while self.looping:
      last_loop = int(time.time())
      if not self.stopped:
        self.reset()
        self.startup()
        self.close()
        self.write('status_msg: Disabled\nstatus_tag: idle\n')
        self.pk = None
      if last_loop == int(time.time()):
        time.sleep(1)

  def startup(self):
    raise Exception('Unimplemented')

  def postpone(self, func, argument):
    return func(argument)

  def stop(self, then=False):
    self.stopped = True
    if self.pk:
      self.send('exit: stopping\n')
      self.postpone(self.stop, then)
    else:
      if then:
        then()

  def restart(self):
    self.stopped = False

  def toggle(self):
    if self.stopped:
      self.restart()
    else:
      self.stop()

  def quit(self):
    self.looping = False
    self.stopped = True
    if self.pk:
      self.send('exit: quitting\n')
    self.close()
    self.pk = None
    self.join()


class CommThread(threading.Thread):
  daemon = True

  def __init__(self, pkThread):
    threading.Thread.__init__(self)
    self.pkThread = pkThread
    self.looping = False

    self.multi = None
    self.multi_args = None

    # Callbacks
    self.cb = {}

  def call_cb(self, which, args):
    return self.cb[which](args)

  def parse_line(self, line):
#   print '<< %s' % line[:-1]
    if line.startswith('begin_'):
      self.multi = line[6:].strip()
      self.multi_args = {'_raw': []}
    elif self.multi:
      if line.startswith('end_'):
        if self.multi in self.cb:
          self.call_cb(self.multi, self.multi_args)
        elif 'default' in self.multi_args:
          self.pkThread.send(self.multi_args['default']+'\n')
        self.multi = self.multi_args = None
      else:
        self.multi_args['_raw'].append(line.strip())
        try:
          variable, value = line.strip().split(': ', 1)
          self.multi_args[variable] = value
        except ValueError:
          pass
    else:
      try:
        command, args = line.strip().split(': ', 1)
        if command in self.cb:
          self.call_cb(command, args)
      except ValueError:
        pass

  def run(self):
    self.pkThread.start()
    self.looping = True
    line = ''
    while self.looping:
      line += self.pkThread.recv(1)
      if line.endswith('\n'):
        self.parse_line(line)
        line = ''

  def quit(self):
    self.pkThread.quit()
    self.looping = False
